(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE CommonSettings;

        (****************************************************************)
        (*                                                              *)
        (*                  Setup program for SFtpServer                *)
        (*                   Data common to the program                 *)
        (*                                                              *)
        (*    Last edited:    28 July 2022                              *)
        (*    Status:         OK                                        *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT CARD8;

IMPORT OS2, Strings, FileSys, INIData, Languages;

FROM Names IMPORT
    (* type *)  FilenameString;

FROM LowLevel IMPORT
    (* proc *)  IANDB;

(************************************************************************)

CONST
    Nul = CHR(0);
    SetupINIname0 = "Setup";

VAR
    OurFontName: FontName;
    SetupINIname: ARRAY [0..127] OF CHAR;
    UseTNI: BOOLEAN;
    OurLanguage: Languages.LangHandle;
    LanguageCode: ARRAY [0..31] OF CHAR;

TYPE
    WindowPosition = RECORD
                         x, y: INTEGER;
                     END (*RECORD*);

(************************************************************************)

PROCEDURE SetINImode (textmode: BOOLEAN);

    (* Records whether to use a TNI file for the Setup data.  *)

    BEGIN
        UseTNI := textmode;
        SetupINIname := SetupINIname0;
        IF UseTNI THEN
            Strings.Append (".TNI", SetupINIname);
        ELSE
            Strings.Append (".INI", SetupINIname);
        END (*IF*);
    END SetINImode;

(************************************************************************)

PROCEDURE GetINIname (VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the name of the Setup INI file.  *)

    BEGIN
        Strings.Assign (SetupINIname, name);
    END GetINIname;

(************************************************************************)

PROCEDURE WhichServers (VAR (*OUT*) FTP, SFTP: BOOLEAN;
                                    VAR (*OUT*) string: ARRAY OF CHAR);

    (* Returns flags for which server(s) we are configuring, and sets   *)
    (* string to a textual form of the same information.                *)

    VAR app: ARRAY [0..15] OF CHAR;
        hini: INIData.HINI;
        server: CARD8;

    BEGIN
        hini := INIData.OpenINIFile(SetupINIname);
        IF INIData.INIValid(hini) THEN
            app := "Setup";
            IF NOT INIData.INIGet (hini, app, "server", server) THEN
                server := 4;
            END (*IF*);
            INIData.CloseINIFile (hini);
        ELSE
            server := 4;
        END (*IF*);
        FTP := ODD (server);
        SFTP := IANDB (server, 4) <> 0;
        IF FTP THEN
            Strings.Assign ("FTP", string);
            IF SFTP THEN
                Strings.Append (" & SFTP", string);
            END (*IF*);
        ELSIF SFTP THEN
            Strings.Assign ("SFTP", string);
        ELSE
            Strings.Assign ("??", string);
        END (*IF*);
    END WhichServers;

(************************************************************************)

PROCEDURE CurrentFont (VAR (*OUT*) fontname: FontName);

    (* Returns the currently set font for the notebook. *)

    BEGIN
        fontname:= OurFontName;
    END CurrentFont;

(************************************************************************)

PROCEDURE UpdateFontFrom (hwnd: OS2.HWND);

    (* Takes the font setting from window hwnd and propagates it to the *)
    (* entire notebook.  Note that this will often be a "no operation"  *)
    (* because the font is no different from the presently active one.  *)

    VAR NewFontName: FontName;
        AttrFound, length: CARDINAL;
        hini: INIData.HINI;  target: OS2.HWND;
        app: ARRAY [0..4] OF CHAR;

    BEGIN
        length := OS2.WinQueryPresParam (hwnd, OS2.PP_FONTNAMESIZE, 0,
                                     AttrFound, FontNameSize, NewFontName,
                                      0(*OS2.QPF_NOINHERIT*));
        IF length < FontNameSize THEN
            NewFontName[length] := Nul;
        END (*IF*);

        IF NOT Strings.Equal (NewFontName, OurFontName) THEN

            OurFontName := NewFontName;
            hini := INIData.OpenINIFile (SetupINIname);
            app := "Font";
            INIData.INIPutString (hini, app, "MainNotebook", OurFontName);
            INIData.CloseINIFile (hini);

            (* For reasons that are still a mystery to me, we have to go    *)
            (* up three levels in the hierarchy to get from here to the     *)
            (* frame.  (I calculated two, but there's apparently an extra   *)
            (* container window that we're not told about.)                 *)

            target := OS2.WinQueryWindow(hwnd, OS2.QW_PARENT);
            target := OS2.WinQueryWindow(target, OS2.QW_PARENT);
            target := OS2.WinQueryWindow(target, OS2.QW_PARENT);
            OS2.WinSendMsg (target, FONTCHANGED, NIL, NIL);

        END (*IF*);

    END UpdateFontFrom;

(************************************************************************)

PROCEDURE SetDefaultFont;

    (* Sets initial font if one isn't stored. *)

    VAR hini: INIData.HINI;
        app: ARRAY [0..4] OF CHAR;

    BEGIN
        OurFontName := "";
        hini := INIData.OpenINIFile(SetupINIname);
        IF INIData.INIValid(hini) THEN
            app := "Font";
            IF NOT INIData.INIGetString (hini, app, "MainNotebook", OurFontName)
                          OR (OurFontName[0] = Nul) THEN
                OurFontName := "10.System Proportional";
                INIData.INIPutString (hini, app, "MainNotebook", OurFontName);
            END (*IF*);
            INIData.CloseINIFile (hini);
        ELSE
            OurFontName := "10.System Proportional";
        END (*IF*);
    END SetDefaultFont;

(************************************************************************)

PROCEDURE ChangeLanguageTo (name: ARRAY OF CHAR): BOOLEAN;

    (* Changes the current language setting, but only if a suitable     *)
    (* setup.name.lng file exists.  Returns FALSE if no change.         *)

    CONST prefix = "Setup";

    VAR result: BOOLEAN;  file: FilenameString;

    BEGIN
        Strings.Capitalize (name);
        result := FALSE;
        IF NOT Strings.Equal (name, LanguageCode) THEN
            Strings.Assign (prefix, file);
            Strings.Append (".", file);
            Strings.Append (name, file);
            Strings.Append (".lng", file);
            IF FileSys.Exists (file) THEN
                Languages.DropLanguage (OurLanguage);
                Strings.Assign (name, LanguageCode);
                OurLanguage := Languages.UseLanguage (prefix, name);
                result := TRUE;
            END (*IF*);
        END (*IF*);
        RETURN result;
    END ChangeLanguageTo;

(************************************************************************)

PROCEDURE CurrentLanguage (VAR (*OUT*) lang: Languages.LangHandle;
                                VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the current language setting. *)

    BEGIN
        lang := OurLanguage;
        Strings.Assign (LanguageCode, name);
    END CurrentLanguage;

(************************************************************************)

PROCEDURE SetInitialLanguage;

    (* Sets the language from the SETUP INI or TNI file. *)

    VAR app: ARRAY [0..7] OF CHAR;
        hini: INIData.HINI;

    BEGIN
        app := 'Setup';
        hini := INIData.OpenINIFile(SetupINIname);
        IF NOT INIData.INIValid (hini)
           OR NOT INIData.INIGetString (hini, app, 'Language', LanguageCode)
                                     OR (LanguageCode[0] = Nul) THEN
            LanguageCode := "en";
        END (*IF*);
        OurLanguage := Languages.UseLanguage ("Setup", LanguageCode);
        INIData.CloseINIFile (hini);
    END SetInitialLanguage;

(************************************************************************)

PROCEDURE SetInitialWindowPosition (hwnd: OS2.HWND;  label: ARRAY OF CHAR);

    (* If this window has a previously stored position in our INI file, *)
    (* positions the window to that position.  Ditto for font.          *)

    CONST bufsize = 256;

    VAR hini: INIData.HINI;  pos: WindowPosition;
        FontName: ARRAY [0..bufsize-1] OF CHAR;
        app: ARRAY [0..9] OF CHAR;

    BEGIN
        hini := INIData.OpenINIFile (SetupINIname);
        IF INIData.INIValid(hini) THEN
            app := "WindowPos";
            IF INIData.INIGet (hini, app, label, pos) THEN
                OS2.WinSetWindowPos (hwnd, 0, pos.x, pos.y, 0, 0, OS2.SWP_MOVE);
            END (*IF*);
            app := "Font";
            IF NOT INIData.INIGetString (hini, app, label, FontName)
                          OR (FontName[0] = Nul) THEN
                FontName := "8.Helv";
            END (*IF*);
            INIData.CloseINIFile (hini);
        ELSE
            FontName := "8.Helv";
        END (*IF*);
        OS2.WinSetPresParam (hwnd, OS2.PP_FONTNAMESIZE, bufsize, FontName);
    END SetInitialWindowPosition;

(************************************************************************)

PROCEDURE StoreWindowPosition (hwnd: OS2.HWND;  label: ARRAY OF CHAR;
                                                storefont: BOOLEAN);

    (* Saves the location and (optionally) font of this window in our INI file. *)

    CONST bufsize = 256;

    VAR hini: INIData.HINI;  swp: OS2.SWP;
        pos: WindowPosition;
        AttrFound, length: CARDINAL;
        FontName: ARRAY [0..bufsize-1] OF CHAR;
        app: ARRAY [0..10] OF CHAR;

    BEGIN
        OS2.WinQueryWindowPos (hwnd, swp);
        pos.x := swp.x;  pos.y := swp.y;
        IF storefont THEN
            length := OS2.WinQueryPresParam (hwnd, OS2.PP_FONTNAMESIZE, 0,
                                         AttrFound,
                                         bufsize, FontName, 0(*OS2.QPF_NOINHERIT*));
            IF length < bufsize THEN
                FontName[length] := Nul;
            END (*IF*);
        END (*IF*);
        hini := INIData.OpenINIFile (SetupINIname);
        IF INIData.INIValid(hini) THEN
            app := "WindowPos";
            INIData.INIPut (hini, app, label, pos);
            IF storefont THEN
                app := "Font";
                INIData.INIPutString (hini, app, label, FontName);
            END (*IF*);
            INIData.CloseINIFile (hini);
        END (*IF*);
    END StoreWindowPosition;

(************************************************************************)

END CommonSettings.

