(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE EditUser;

        (************************************************************)
        (*                                                          *)
        (*                  PM Setup for SFtpServer                 *)
        (*          Dialogue to edit the data for one user          *)
        (*                                                          *)
        (*    Last edited:    15 September 2022                     *)
        (*    Status:         OK                                    *)
        (*                                                          *)
        (************************************************************)


IMPORT SYSTEM, OS2, CommonSettings, DID, Strings, Tree, UIPControls, IPFilters, HideList;

FROM Arith64 IMPORT
    (* const*)  Zero64, Max64,
    (* type *)  CARD64,
    (* proc *)  Compare64;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer, StrToBufferA;

FROM MiscFuncs IMPORT
    (* proc *)  ToLower, EVAL;

FROM Misc IMPORT
    (* proc *)  WinSetDlgItemCard, WinQueryDlgItemCard;

FROM MD5 IMPORT
    (* type *)  MD5_CTX, MD5_DigestType,
    (* proc *)  MD5Init, MD5Update, MD5Final, MD5DigestToString;

FROM FSUINI IMPORT
    (* type *)  FHINI,
    (* proc *)  OpenINIFiles, OpenINIForUser, CloseINIFiles,
                ItemSize, INIFetch, INIGetCard, INIGetString,
                INIDeleteApp, INIDeleteKey, INIPutString, INIPut;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(**************************************************************************)

CONST
    Nul = CHR(0);
    NameLength = 256;
    RealNameSize = 64;
    PasswordSize = 32;
    NoteFieldSize = 2048;
    MEG = 1048576;

TYPE
    UserCategory = (NoSuchUser, NoPasswordNeeded, GuestUser, NormalUser,
                    Manager, UserTemplate);
    NameString = ARRAY [0..NameLength-1] OF CHAR;
    PassString = ARRAY [0..PasswordSize] OF CHAR;

VAR
    pagehandle: OS2.HWND;
    ourlang: LangHandle;
    configFTP, configSFTP: BOOLEAN;

    User: RECORD
              LoginLimit: CARDINAL;
              useFTP, useSFTP, allow3way, NewUser: BOOLEAN;
              NameChanged, SingleUse, DeleteExpired, SuppressLog: BOOLEAN;
              UseTemplate, OldUseTemplate: BOOLEAN;
              PasswordChanged, EncryptPassword,
                         OriginalEncryptPassword : BOOLEAN;
              category: UserCategory;
              name, OldName, TemplateName: NameString;
              Password, OriginalPassword: PassString;
              RealName: ARRAY [0..RealNameSize-1] OF CHAR;
              Notes: ARRAY [0..NoteFieldSize-1] OF CHAR;
              limit, speedlimit, upspeedlimit: CARDINAL;
              upsizelimit: CARD64;
              UserTreeRoot, TemplateTreeRoot: Tree.DirEntryPtr;
              UIPList: IPFilters.ListPtr;
              HideList: HideList.ListPtr;
          END (*RECORD*);

    swp: ARRAY [0..3] OF OS2.SWP;
        (* These are positions for the two variable-position subwindows. *)

    HidePasswords, DetectLLChanges: BOOLEAN;
    passwindow: OS2.HWND;

(************************************************************************)
(*           LABELLING DIALOGUE ITEMS IN THE CURRENT LANGUAGE           *)
(************************************************************************)

PROCEDURE SetLanguage (hwnd: OS2.HWND;  lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    (********************************************************************)

    PROCEDURE SetOne (item: ARRAY OF CHAR;  place: CARDINAL);

        BEGIN
            StrToBuffer (lang, item, stringval);
            OS2.WinSetDlgItemText (hwnd, place, stringval);
        END SetOne;

    (********************************************************************)

    BEGIN
        ourlang := lang;
        StrToBufferA (lang, "EditUser.Title", User.name, stringval);
        OS2.WinSetWindowText (hwnd, stringval);
        SetOne ("Usertype.nopassword", DID.NoPassword);
        SetOne ("Usertype.guest", DID.Guest);
        SetOne ("Usertype.user", DID.User);
        SetOne ("Usertype.manager", DID.Manager);
        SetOne ("Usertype.template", DID.IsTemplate);

        SetOne ("Edituser.allow3way", DID.allow3way);
        SetOne ("Edituser.lastlogin", DID.LastLoginLabel);
        SetOne ("Edituser.addIPcontrols", DID.EditUIPListButton);
        SetOne ("Edituser.addHideList", DID.EditHideListButton);
        SetOne ("Edituser.useFTP", DID.UseFTP);
        SetOne ("Edituser.useSFTP", DID.UseSFTP);

        SetOne ("Edituser.speedlimit", DID.SpeedLimitGroup);
        SetOne ("Edituser.download", DID.SpeedLimitLabel1);
        SetOne ("Edituser.bytespersec", DID.SpeedLimitLabel2);
        SetOne ("Edituser.upload", DID.SpeedLimitLabel3);
        SetOne ("Edituser.bytespersec", DID.SpeedLimitLabel4);
        SetOne ("Edituser.sizelimit", DID.UpSizeLimitLabel);
        SetOne ("Edituser.MiB", DID.MiB);
        SetOne ("Edituser.userlimit", DID.UserLimitLabel);
        SetOne ("Edituser.RealName", DID.RealNameLabel);
        SetOne ("Edituser.Username", DID.UsernameLabel);
        SetOne ("Edituser.Password", DID.PasswordLabel);
        SetOne ("Edituser.encryptpassword", DID.EncryptPassword);
        SetOne ("Edituser.suppresslog", DID.SuppressLog);

        SetOne ("Edituser.limitlogins", DID.LimitLogins);
        SetOne ("Edituser.deleteafter", DID.limit_delete);
        SetOne ("Edituser.logins", DID.TextLogins);
        SetOne ("Edituser.deactivateafter", DID.limit_deactivate);
        SetOne ("Edituser.logins", DID.TextLogins2);

        SetOne ("Edituser.NOTES", DID.NOTES);

        SetOne ("Buttons.OK", DID.UserOK);
        SetOne ("Buttons.Cancel", DID.UserCancel);

        SetOne ("Edituser.EditDirs", DID.EditDirButton);
        SetOne ("Edituser.usetemplate", DID.UseTemplate);

    END SetLanguage;

(************************************************************************)
(*                   ENABLING AND DISABLING FIELDS                      *)
(************************************************************************)

PROCEDURE EnableFields (hwnd: OS2.HWND);

    (* Enables or disables some of the dialogue fields, where the       *)
    (* decision depends on whether we're configuring FTP or SFTP.       *)

    (********************************************************************)

    PROCEDURE DoOne (field: CARDINAL;  cond: BOOLEAN);

        VAR w: OS2.HWND;

        BEGIN
            w := OS2.WinWindowFromID (hwnd, field);

            (* I have never discovered why WinEnableWindow won't work   *)
            (* if the second parameter is a variable.                   *)

            IF cond THEN
                OS2.WinShowWindow (w, TRUE);
            ELSE
                OS2.WinShowWindow (w, FALSE);
            END (*IF*);
        END DoOne;

    (********************************************************************)

    VAR FTP: BOOLEAN;

    BEGIN
        FTP := configFTP AND User.useFTP;
        DoOne (DID.allow3way, FTP AND
                ((User.category = NormalUser) OR (User.category = Manager)));
        DoOne (DID.SpeedLimitGroup, FTP);
        DoOne (DID.SpeedLimit, FTP);
        DoOne (DID.UpSpeedLimit, FTP);
        DoOne (DID.SpeedLimitLabel1, FTP);
        DoOne (DID.SpeedLimitLabel2, FTP);
        DoOne (DID.SpeedLimitLabel3, FTP);
        DoOne (DID.SpeedLimitLabel4, FTP);
        DoOne (DID.UpSizeLimitLabel, FTP);
        DoOne (DID.UpSizeLimit, FTP);
        DoOne (DID.MiB, FTP);
        (*
        DoOne (DID.UserLimitLabel, FTP);
        DoOne (DID.UserLimit, FTP);
        *)
        DoOne (DID.SuppressLog, FTP);
    END EnableFields;

(************************************************************************)

PROCEDURE SetServer (setFTP, setSFTP: BOOLEAN);

    (* Specifies which server(s) we are configuring. *)

    BEGIN
        configFTP := setFTP;
        configSFTP := setSFTP;
    END SetServer;

(************************************************************************)
(*                ENABLING FIELDS RELATED TO LOGIN LIMIT                *)
(************************************************************************)

PROCEDURE SetLoginLimit (hwnd: OS2.HWND;  limit: CARDINAL);

    (* Saves limit in User.LoginLimit.  Shows or hides some dialogue    *)
    (* elements depending on whether it is zero or nonzero.             *)

    VAR w1, w2, w3, w5, w6: OS2.HWND;

    BEGIN
        DetectLLChanges := FALSE;
        User.LoginLimit := limit;
        OS2.WinSendDlgItemMsg (hwnd, DID.LimitLogins, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(limit>0)), NIL);
        WinSetDlgItemCard (hwnd, DID.LoginLimit, limit);

        w1 := OS2.WinWindowFromID(hwnd, DID.LimitLogins);
        w2 := OS2.WinWindowFromID(hwnd, DID.LoginLimit);
        w3 := OS2.WinWindowFromID(hwnd, DID.TextLogins);
        w5 := OS2.WinWindowFromID(hwnd, DID.limit_delete);
        w6 := OS2.WinWindowFromID(hwnd, DID.limit_deactivate);

        IF User.DeleteExpired THEN
            OS2.WinSendDlgItemMsg (hwnd, DID.limit_delete, OS2.BM_SETCHECK,
                                         OS2.MPFROMSHORT(1), NIL);
            OS2.WinSendDlgItemMsg (hwnd, DID.limit_deactivate, OS2.BM_SETCHECK,
                                         OS2.MPFROMSHORT(0), NIL);
            OS2.WinSetWindowPos (w2, 0, swp[0].x, swp[0].y, 0, 0, OS2.SWP_MOVE);
            OS2.WinSetWindowPos (w3, 0, swp[1].x, swp[1].y, 0, 0, OS2.SWP_MOVE);
        ELSE
            OS2.WinSendDlgItemMsg (hwnd, DID.limit_delete, OS2.BM_SETCHECK,
                                         OS2.MPFROMSHORT(0), NIL);
            OS2.WinSendDlgItemMsg (hwnd, DID.limit_deactivate, OS2.BM_SETCHECK,
                                         OS2.MPFROMSHORT(1), NIL);
            OS2.WinSetWindowPos (w2, 0, swp[2].x, swp[2].y, 0, 0, OS2.SWP_MOVE);
            OS2.WinSetWindowPos (w3, 0, swp[3].x, swp[3].y, 0, 0, OS2.SWP_MOVE);
        END (*IF*);

        IF limit = 0 THEN
            OS2.WinEnableWindow (w2, FALSE);
            OS2.WinEnableWindow (w3, FALSE);
            OS2.WinEnableWindow (w5, FALSE);
            OS2.WinEnableWindow (w6, FALSE);
        ELSE
            OS2.WinEnableWindow (w2, TRUE);
            OS2.WinEnableWindow (w3, TRUE);
            OS2.WinEnableWindow (w5, TRUE);
            OS2.WinEnableWindow (w6, TRUE);
            IF limit = 1 THEN
                OS2.WinSetWindowText (w3, "login");
            ELSE
                OS2.WinSetWindowText (w3, "logins");
            END (*IF*);
        END (*IF*);
        DetectLLChanges := TRUE;

    END SetLoginLimit;

(************************************************************************)
(*               TURNING 'TEMPLATE' OPTIONS ON AND OFF                  *)
(************************************************************************)

PROCEDURE EnableUseTemplateFields (hwnd: OS2.HWND;  enable: BOOLEAN);

    (* Enables or disables the dialogue elements associated with the    *)
    (* "use a template" option.                                         *)

    VAR subwindow1, subwindow2: OS2.HWND;
        root: Tree.DirEntryPtr;

    BEGIN
        subwindow1 := OS2.WinWindowFromID(hwnd, DID.TemplateName);
        subwindow2 := OS2.WinWindowFromID(hwnd, DID.EditDirButton);
        IF enable THEN
             OS2.WinEnableWindow (subwindow1, TRUE);
             OS2.WinEnableWindow (subwindow2, FALSE);
             root := User.TemplateTreeRoot;
        ELSE
             OS2.WinEnableWindow (subwindow1, FALSE);
             OS2.WinEnableWindow (subwindow2, TRUE);
             root := User.UserTreeRoot;
        END (*IF*);
        User.UseTemplate := enable;
        User.OldUseTemplate := enable;
        Tree.InitialDisplay (OS2.WinWindowFromID(hwnd, DID.DirSummary), root);
    END EnableUseTemplateFields;

(************************************************************************)

PROCEDURE EnableIsTemplateFields (hwnd: OS2.HWND;  enable: BOOLEAN);

    (* Enables or disables the dialogue elements associated with the    *)
    (* "is a template" option.                                          *)

    VAR title: ARRAY [0..NameLength-1] OF CHAR;
        stringval: ARRAY [0..63] OF CHAR;
        UseTemplateWasOn: BOOLEAN;

    BEGIN
        IF enable THEN

            IF User.OldName[0] = Nul THEN
                StrToBuffer (ourlang, "Edituser.newtemplate", stringval);
                title := "New template";
            ELSE
                StrToBufferA (ourlang, "Edituser.titletemplate", User.OldName, stringval);
                title := "Template name: ";
            END (*IF*);

            (* Turn off the "use template" option. *)

            UseTemplateWasOn := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.UseTemplate,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
            OS2.WinSendDlgItemMsg (hwnd, DID.UseTemplate, OS2.BM_SETCHECK,
                             OS2.MPFROMSHORT(ORD(FALSE)), NIL);
            EnableUseTemplateFields (hwnd, FALSE);
            User.OldUseTemplate := UseTemplateWasOn;
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UseTemplate), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.TemplateName), FALSE);

            (* Turn off a bunch of other things. *)

            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LastLoginLabel), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LastLogin), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LimitLogins), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.limit_delete), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.limit_deactivate), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LoginLimit), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.TextLogins), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SuppressLog), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel1), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel2), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel3), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel4), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitGroup), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimit), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UpSpeedLimit), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UserLimitLabel), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UserLimit), FALSE);
            OS2.WinSetWindowPos (OS2.WinWindowFromID(hwnd, DID.RealNameLabel),
                                 0, 0, 0, 150, 25, OS2.SWP_SIZE);
            StrToBuffer (ourlang, "Edituser.templatename", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.RealNameLabel, stringval);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.RealName), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UsernameLabel), FALSE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.PasswordLabel), FALSE);
            OS2.WinShowWindow (passwindow, FALSE);

        ELSE
            IF User.OldName[0] = Nul THEN
                StrToBuffer (ourlang, "Edituser.newuser", title);
            ELSE
                StrToBufferA (ourlang, "Edituser.Title", User.OldName, title);
            END (*IF*);

            IF User.OldUseTemplate THEN
                OS2.WinSendDlgItemMsg (hwnd, DID.UseTemplate, OS2.BM_SETCHECK,
                             OS2.MPFROMSHORT(ORD(TRUE)), NIL);
            END (*IF*);
            EnableUseTemplateFields (hwnd, User.OldUseTemplate);

            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LastLoginLabel), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LastLogin), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LimitLogins), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.limit_delete), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.limit_deactivate), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.LoginLimit), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.TextLogins), TRUE);
            SetLoginLimit (hwnd, User.LoginLimit);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SuppressLog), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel1), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel2), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel3), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitLabel4), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimitGroup), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.SpeedLimit), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UpSpeedLimit), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UserLimitLabel), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UserLimit), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.RealName), TRUE);
            OS2.WinSetWindowPos (OS2.WinWindowFromID(hwnd, DID.RealNameLabel),
                                 0, 0, 0, 92, 25, OS2.SWP_SIZE);
            StrToBuffer (ourlang, "Edituser.RealName", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.RealNameLabel, stringval);

            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UsernameLabel), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.PasswordLabel), TRUE);
            OS2.WinShowWindow (passwindow, TRUE);

            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.UseTemplate), TRUE);
            OS2.WinShowWindow (OS2.WinWindowFromID(hwnd, DID.TemplateName), TRUE);

        END (*IF*);

        OS2.WinSetWindowText (hwnd, title);

    END EnableIsTemplateFields;

(************************************************************************)
(*                     INITIAL LOAD FROM INI FILE                       *)
(************************************************************************)

PROCEDURE LoadDialogueData (hwnd: OS2.HWND;  username: ARRAY OF CHAR;
                                        lang: LangHandle;  clone: BOOLEAN);

    (* Fills in all fields for the dialogue. *)

    VAR val: CARDINAL;
        fhini: FHINI;
        SingleUse: BOOLEAN;
        passstring: PassString;
        stringval: ARRAY [0..31] OF CHAR;
        subwindow1, subwindow2, subwindow3: OS2.HWND;

    BEGIN
        DetectLLChanges := FALSE;
        Strings.Assign (username, User.name);
        fhini := OpenINIFiles();
        IF NOT INIFetch (fhini, '$SYS', "HidePasswords", HidePasswords) THEN
            HidePasswords := FALSE;
        END (*IF*);
        CloseINIFiles (fhini);

        StrToBuffer (lang, "Edituser.lastlogin", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.LastLoginLabel, stringval);
        StrToBuffer (lang, "SortUsers.never", passstring);
        OS2.WinSetDlgItemText (hwnd, DID.LastLogin, passstring);

        OS2.WinSetDlgItemText (hwnd, DID.UserName, username);
        subwindow1 := OS2.WinWindowFromID(hwnd, DID.Password);
        subwindow2 := OS2.WinWindowFromID(hwnd, DID.PasswordH);
        IF HidePasswords THEN
            OS2.WinShowWindow (subwindow1, FALSE);
            passwindow := subwindow2;
        ELSE
            OS2.WinShowWindow (subwindow2, FALSE);
            passwindow := subwindow1;
        END (*IF*);
        OS2.WinShowWindow (passwindow, FALSE);

        User.useFTP := TRUE;
        User.useSFTP := TRUE;
        User.allow3way := FALSE;
        User.NewUser := username[0] = Nul;
        User.SingleUse := FALSE;
        User.DeleteExpired := FALSE;
        User.SuppressLog := FALSE;
        User.LoginLimit := 0;
        User.Password := "";
        User.RealName := "";
        User.Notes := "";
        User.UserTreeRoot := NIL;
        User.TemplateTreeRoot := NIL;
        User.UseTemplate := FALSE;
        User.TemplateName := "";
        User.UIPList := NIL;

        IF User.NewUser THEN

            User.OriginalEncryptPassword := FALSE;
            User.EncryptPassword := HidePasswords;
            User.category := NormalUser;
            User.useFTP := TRUE;
            User.useSFTP := TRUE;

            (* Limits. *)

            val := MAX(CARDINAL);
            WinSetDlgItemCard (hwnd, DID.UserLimit, val);
            WinSetDlgItemCard (hwnd, DID.SpeedLimit, val);
            WinSetDlgItemCard (hwnd, DID.UpSpeedLimit, val);
            StrToBuffer (lang, "Sortusers.unlimited", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.UpSizeLimit, stringval);

            fhini := OpenINIFiles();

        ELSE

            fhini := OpenINIForUser(username);

            IF NOT INIGetString (fhini, username, "LastLogin", passstring) THEN
                StrToBuffer (lang, "SortUsers.never", passstring);
            END (*IF*);
            OS2.WinSetDlgItemText (hwnd, DID.LastLogin, passstring);

            IF NOT INIFetch (fhini, username, "Active", User.useFTP) THEN
                User.useFTP := FALSE;
            END (*IF*);

            IF NOT INIFetch (fhini, username, "useSFTP", User.useSFTP) THEN
                User.useSFTP := User.useFTP;
            END (*IF*);

            IF NOT INIFetch (fhini, username, "allow3way", User.allow3way) THEN
                User.allow3way := FALSE;
            END (*IF*);

            IF INIFetch (fhini, username, "SingleUse", SingleUse) THEN
                (* Obsolete option *)
                IF SingleUse THEN
                    SetLoginLimit (hwnd, 1);
                ELSE
                    SetLoginLimit (hwnd, 0);
                END (*IF*);
                INIDeleteKey (fhini, username, "SingleUse");
            END (*IF*);

            EVAL (INIFetch (fhini, username, "LoginLimit", User.LoginLimit));
            EVAL (INIFetch (fhini, username, "DeleteExpired", User.DeleteExpired));
            EVAL (INIFetch (fhini, username, "SuppressLog", User.SuppressLog));

            IF INIFetch (fhini, username, "EncryptPassword", User.EncryptPassword) THEN
                User.OriginalEncryptPassword := User.EncryptPassword;
            ELSE
                User.OriginalEncryptPassword := FALSE;
                User.EncryptPassword := HidePasswords;
            END (*IF*);

            EVAL (INIGetString (fhini, username, "Password", User.Password));
            EVAL (INIGetString (fhini, username, "RealName", User.RealName));
            EVAL (INIGetString (fhini, username, "Notes", User.Notes));

            IF (NOT INIFetch (fhini, username, "Category", User.category))
                           OR (User.category = NoSuchUser) THEN
                User.category := NormalUser;
            END (*IF*);

            EVAL (INIFetch (fhini, username, "UseTemplate", User.UseTemplate));
            EVAL (INIGetString (fhini, username, "TemplateName", User.TemplateName));

            (* User limit and speed limits. *)

            IF NOT INIGetCard (fhini, username, "UserLimit", val) THEN
                val := MAX(CARDINAL);
            END (*IF*);
            WinSetDlgItemCard (hwnd, DID.UserLimit, val);

            IF NOT INIGetCard (fhini, username, "SpeedLimit", val) THEN
                val := MAX(CARDINAL);
            ELSIF val = 0 THEN
                val := 1;
            END (*IF*);
            WinSetDlgItemCard (hwnd, DID.SpeedLimit, val);

            IF NOT INIGetCard (fhini, username, "UpSpeedLimit", val) THEN
                (* Keep previous val. *)
            ELSIF val = 0 THEN
                val := 1;
            END (*IF*);
            WinSetDlgItemCard (hwnd, DID.UpSpeedLimit, val);

            (* UpSizeLimit is a special case.  We store it in the INI   *)
            (* file as a CARD64, but for display we divide it by 2^20   *)

            IF NOT INIFetch (fhini, username, "UpSizeLimit", User.upsizelimit) THEN
                User.upsizelimit := Zero64;
            END (*IF*);
            IF (Compare64(User.upsizelimit,Zero64)=0) OR (User.upsizelimit.high > 4095) THEN
                StrToBuffer (lang, "Edituser.unlimited", stringval);
                OS2.WinSetDlgItemText (hwnd, DID.UpSizeLimit, stringval);
            ELSE
                (* Divide by 2^20. *)
                val := 4096*User.upsizelimit.high + User.upsizelimit.low / MEG;
                WinSetDlgItemCard (hwnd, DID.UpSizeLimit, val);
            END (*IF*);

        END (*IF*);

        User.TemplateTreeRoot := Tree.LoadPermissions (fhini, User.TemplateName);
        User.UserTreeRoot := Tree.LoadPermissions (fhini, username);

        CloseINIFiles (fhini);

        User.UIPList := IPFilters.LoadIPFilterList (username, "IPfilter");
        IF (User.UIPList = NIL) OR (User.UIPList^.next = NIL) THEN
            StrToBuffer (lang, "Edituser.addIPcontrols", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.EditUIPListButton, stringval);
        ELSE
            StrToBuffer (lang, "Edituser.modifyIPcontrols", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.EditUIPListButton, stringval);
        END (*IF*);

        User.HideList := HideList.LoadList (username);
        IF (User.HideList = NIL) OR (User.HideList^.next = NIL) THEN
            StrToBuffer (lang, "Edituser.addHideList", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.EditHideListButton, stringval);
        ELSE
            StrToBuffer (lang, "Edituser.modifyHideList", stringval);
            OS2.WinSetDlgItemText (hwnd, DID.EditHideListButton, stringval);
        END (*IF*);

        User.OldUseTemplate := User.UseTemplate;
        User.OriginalPassword := User.Password;
        User.PasswordChanged := FALSE;
        SetLoginLimit (hwnd, User.LoginLimit);

        OS2.WinSendDlgItemMsg (hwnd, DID.UseFTP, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(User.useFTP)), NIL);
        OS2.WinSendDlgItemMsg (hwnd, DID.UseSFTP, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(User.useSFTP)), NIL);
        OS2.WinSendDlgItemMsg (hwnd, DID.allow3way, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(User.allow3way)), NIL);
        OS2.WinSendDlgItemMsg (hwnd, DID.SuppressLog, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(User.SuppressLog)), NIL);
        OS2.WinSendDlgItemMsg (hwnd, DID.EncryptPassword, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(ORD(User.EncryptPassword)), NIL);

        passstring := User.Password;
        CASE User.category OF
          |  NoSuchUser:
          |  NoPasswordNeeded:
                OS2.WinSendDlgItemMsg (hwnd, DID.NoPassword, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(1), NIL);
                passstring := "";
          |  GuestUser:
                OS2.WinSendDlgItemMsg (hwnd, DID.Guest, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(1), NIL);
                passstring := "@";
          |  NormalUser:
                OS2.WinSendDlgItemMsg (hwnd, DID.User, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(1), NIL);
          |  Manager:
                OS2.WinSendDlgItemMsg (hwnd, DID.Manager, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(1), NIL);
          |  UserTemplate:
                OS2.WinSendDlgItemMsg (hwnd, DID.IsTemplate, OS2.BM_SETCHECK,
                                 OS2.MPFROMSHORT(1), NIL);
                passstring := "";
        END (*CASE*);

        subwindow1 := passwindow;
        subwindow2 := OS2.WinWindowFromID(hwnd, DID.EncryptPassword);
        subwindow3 := OS2.WinWindowFromID(hwnd, DID.allow3way);
        IF (User.category = NoPasswordNeeded)
                       OR (User.category = GuestUser)
                       OR (User.category = UserTemplate) THEN
             OS2.WinEnableWindow (subwindow1, FALSE);
             OS2.WinShowWindow (subwindow2, FALSE);
             OS2.WinShowWindow (subwindow3, FALSE);
        ELSE
             OS2.WinEnableWindow (subwindow1, TRUE);
             OS2.WinShowWindow (subwindow2, TRUE);
             OS2.WinShowWindow (subwindow3, TRUE);
        END (*IF*);

        OS2.WinSetWindowText (passwindow, passstring);
        OS2.WinSetDlgItemText (hwnd, DID.RealName, User.RealName);
        OS2.WinSetDlgItemText (hwnd, DID.UserNotes, User.Notes);

        OS2.WinSendDlgItemMsg (hwnd, DID.UseTemplate, OS2.BM_SETCHECK,
                             OS2.MPFROMSHORT(ORD(User.UseTemplate)), NIL);
        OS2.WinSetDlgItemText (hwnd, DID.TemplateName, User.TemplateName);

        EnableUseTemplateFields (hwnd, User.UseTemplate);
        EnableIsTemplateFields (hwnd, User.category = UserTemplate);
        IF clone THEN
            Strings.Append ("_1", User.name);
            OS2.WinSetDlgItemText (hwnd, DID.UserName, User.name);
        END (*IF*);

    END LoadDialogueData;

(************************************************************************)
(*                           SAVING THE DATA                            *)
(************************************************************************)

PROCEDURE SaveUserData (hwnd: OS2.HWND);

    (* Copies dialogue data to the User record, but does not yet write  *)
    (* it back to the INI file.                                         *)

    (********************************************************************)

    PROCEDURE CheckButton (button: CARDINAL): BOOLEAN;

        BEGIN
            RETURN OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, button,
                                   OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        END CheckButton;

    (********************************************************************)

    VAR val: CARDINAL;

    BEGIN
        OS2.WinQueryDlgItemText (hwnd, DID.UserName, SIZE(NameString), User.name);

        (* Remove leading spaces from username, and make it lowercase. *)

        WHILE User.name[0] = ' ' DO
            Strings.Delete (User.name, 0, 1);
        END (*WHILE*);
        ToLower (User.name);

        IF CheckButton (DID.NoPassword) THEN
            User.category := NoPasswordNeeded;
        ELSIF CheckButton (DID.Guest) THEN
            User.category := GuestUser;
        ELSIF CheckButton (DID.User) THEN
            User.category := NormalUser;
        ELSIF CheckButton (DID.Manager) THEN
            User.category := Manager;
        ELSIF CheckButton (DID.IsTemplate) THEN
            User.category := UserTemplate;
        END (*IF*);


        User.useFTP := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.UseFTP,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        User.useSFTP := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.UseSFTP,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        IF (User.category = NormalUser) OR (User.category = Manager) THEN
            User.allow3way := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.allow3way,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        ELSE
            User.allow3way := FALSE;
        END (*IF*);
        User.SuppressLog := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.SuppressLog,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        User.UseTemplate := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.UseTemplate,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;

        IF (User.category = NormalUser) OR (User.category = Manager) THEN
            User.EncryptPassword := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.EncryptPassword,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
            OS2.WinQueryWindowText (passwindow, PasswordSize+1, User.Password);
        END (*IF*);

        OS2.WinQueryDlgItemText (hwnd, DID.RealName, RealNameSize, User.RealName);
        OS2.WinQueryDlgItemText (hwnd, DID.UserNotes, NoteFieldSize, User.Notes);
        OS2.WinQueryDlgItemText (hwnd, DID.TemplateName, NameLength, User.TemplateName);
        WinQueryDlgItemCard (hwnd, DID.LoginLimit, User.LoginLimit);
        WinQueryDlgItemCard (hwnd, DID.UserLimit, User.limit);
        WinQueryDlgItemCard (hwnd, DID.SpeedLimit, User.speedlimit);
        WinQueryDlgItemCard (hwnd, DID.UpSpeedLimit, User.upspeedlimit);
        WinQueryDlgItemCard (hwnd, DID.UpSizeLimit, val);
        IF (val = 0) OR (val = MAX(CARDINAL)) THEN
            User.upsizelimit := Max64;
        ELSE
            (* Multiply val by 2^20 to get User.upsizelimit. *)
            User.upsizelimit.high := val / 4096;
            User.upsizelimit.low := (val MOD 4096) * MEG;
        END (*IF*);
        User.PasswordChanged := (User.EncryptPassword <> User.OriginalEncryptPassword)
                                OR NOT Strings.Equal (User.Password, User.OriginalPassword);

    END SaveUserData;

(**************************************************************************)

PROCEDURE CommitChanges;

    (* Copies dialogue data back to the INI file. *)

    VAR ctx: MD5_CTX;  digest: ARRAY [0..15] OF SYSTEM.CARD8;
        fhini: FHINI;
        buffer: PassString;

    BEGIN
        fhini := OpenINIForUser (User.name);
        INIPut (fhini, User.name, "Category", User.category);
        INIPut (fhini, User.name, 'Active', User.useFTP);
        INIPut (fhini, User.name, 'useSFTP', User.useSFTP);
        INIPut (fhini, User.name, 'allow3way', User.allow3way);
        INIPut (fhini, User.name, 'DeleteExpired', User.DeleteExpired);
        INIPut (fhini, User.name, 'LoginLimit', User.LoginLimit);
        INIPut (fhini, User.name, 'SuppressLog', User.SuppressLog);
        INIPut (fhini, User.name, 'UseTemplate', User.UseTemplate);
        INIPut (fhini, User.name, 'EncryptPassword', User.EncryptPassword);
        IF User.PasswordChanged THEN
            buffer := User.Password;
            IF User.EncryptPassword THEN
                ctx := MD5Init();
                MD5Update (ctx, buffer, LENGTH(buffer));
                MD5Final (ctx, digest);
                MD5DigestToString (digest, buffer);
            END (*IF*);
            INIPutString (fhini, User.name, "Password", buffer);
            User.PasswordChanged := FALSE;
        END (*IF*);
        INIPutString (fhini, User.name, "RealName", User.RealName);
        INIPutString (fhini, User.name, "Notes", User.Notes);
        INIPutString (fhini, User.name, "TemplateName", User.TemplateName);
        INIPut (fhini, User.name, "UserLimit", User.limit);
        INIPut (fhini, User.name, "SpeedLimit", User.speedlimit);
        INIPut (fhini, User.name, "UpSpeedLimit", User.upspeedlimit);
        INIPut (fhini, User.name, "UpSizeLimit", User.upsizelimit);
        IF User.UseTemplate THEN
            Tree.SavePermissions (fhini, User.name, NIL);
        ELSE
            Tree.SavePermissions (fhini, User.name, User.UserTreeRoot);
        END (*IF*);
        IPFilters.StoreIPFilterList (fhini, User.name, "IPfilter", User.UIPList);
        HideList.StoreList (fhini, User.name, User.HideList);
        CloseINIFiles (fhini);
    END CommitChanges;

(************************************************************************)

PROCEDURE UserExists (name: ARRAY OF CHAR): BOOLEAN;

    VAR size: CARDINAL;  result: BOOLEAN;
        fhini: FHINI;

    BEGIN
        ToLower (name);
        fhini := OpenINIForUser (name);
        result := ItemSize (fhini, name, "", size) AND (size > 0);
        CloseINIFiles (fhini);
        RETURN result;
    END UserExists;

(************************************************************************)
(*                       THE USER DATA DIALOGUE                         *)
(************************************************************************)

PROCEDURE ["SysCall"] DialogueProc(hwnd     : OS2.HWND
                     ;msg      : OS2.ULONG
                     ;mp1, mp2 : OS2.MPARAM): OS2.MRESULT;

    (* Message handler for the setup dialogue. *)

    TYPE CharSet = SET OF CHAR;

    CONST BoxID = 999;
        Digits = CharSet {'0'..'9'};

    VAR NotificationCode, ButtonID, val: CARDINAL;
        fhini: FHINI;
        PasswordWasDisabled: BOOLEAN;
        encryptwindow, allow3way, w: OS2.HWND;
        Message, stringval: ARRAY [0..255] OF CHAR;
        ButtonValue: BOOLEAN;

    BEGIN
        encryptwindow := OS2.WinWindowFromID(hwnd, DID.EncryptPassword);
        allow3way := OS2.WinWindowFromID(hwnd, DID.allow3way);
        CASE msg OF

              OS2.WM_INITDLG:

                   (* Store the positions of the four moveable subwindows. *)

                   w := OS2.WinWindowFromID (hwnd, DID.LoginLimit);
                   OS2.WinQueryWindowPos (w, swp[0]);
                   w := OS2.WinWindowFromID (hwnd, DID.TextLogins);
                   OS2.WinQueryWindowPos (w, swp[1]);
                   w := OS2.WinWindowFromID (hwnd, DID.LoginLimit2);
                   OS2.WinQueryWindowPos (w, swp[2]);
                   w := OS2.WinWindowFromID (hwnd, DID.TextLogins2);
                   OS2.WinQueryWindowPos (w, swp[3]);

                   RETURN NIL;

           |  OS2.WM_COMMAND:

                   CASE OS2.SHORT1FROMMP(mp1) OF

                     | DID.EditUIPListButton:

                       UIPControls.Edit (hwnd, ourlang, User.UIPList);
                       IF (User.UIPList = NIL) OR
                              ((User.UIPList^.next = NIL) AND User.UIPList^.allow) THEN
                           StrToBuffer (ourlang, "Edituser.addIPcontrols", stringval);
                           OS2.WinSetDlgItemText (hwnd, DID.EditUIPListButton, stringval);
                       ELSE
                           StrToBuffer (ourlang, "Edituser.modifyIPcontrols", stringval);
                           OS2.WinSetDlgItemText (hwnd, DID.EditUIPListButton, stringval);
                       END (*IF*);
                       RETURN NIL;

                     | DID.EditHideListButton:

                       HideList.Edit (hwnd, ourlang, User.HideList);
                       IF (User.HideList = NIL) OR (User.HideList^.next = NIL) THEN
                           StrToBuffer (ourlang, "Edituser.addHideList", stringval);
                           OS2.WinSetDlgItemText (hwnd, DID.EditHideListButton, stringval);
                       ELSE
                           StrToBuffer (ourlang, "Edituser.modifyHideList", stringval);
                           OS2.WinSetDlgItemText (hwnd, DID.EditHideListButton, stringval);
                       END (*IF*);
                       RETURN NIL;

                     | DID.EditDirButton:

                       Tree.Edit (hwnd, User.name, ourlang, User.UserTreeRoot);
                       Tree.InitialDisplay (OS2.WinWindowFromID(hwnd, DID.DirSummary),
                                              User.UserTreeRoot);
                       RETURN NIL;

                     | DID.UserOK:

                       SaveUserData (hwnd);
                       User.NameChanged := NOT Strings.Equal (User.name, User.OldName);
                       IF User.name[0] = Nul THEN
                           StrToBuffer (ourlang, "Edituser.mustbenonblank", Message);
                           OS2.WinMessageBox (OS2.HWND_DESKTOP, hwnd, Message, NIL,
                                      BoxID, OS2.MB_OK + OS2.MB_ERROR);
                           OS2.WinSetFocus (OS2.HWND_DESKTOP, OS2.WinWindowFromID(hwnd, DID.UserName));
                           RETURN NIL;
                       ELSIF NOT User.NameChanged THEN
                           CommitChanges;
                           OS2.WinPostMsg (hwnd, OS2.WM_CLOSE, NIL, NIL);
                           RETURN NIL;
                       END (*IF*);

                       (* If we reach this point then the username has changed.  We have   *)
                       (* to check for duplication of an existing name, and also check     *)
                       (* whether data for the old name should be deleted.                 *)

                       IF UserExists (User.name) THEN

                           StrToBuffer (ourlang, "Edituser.isduplicate", Message);
                           OS2.WinMessageBox (OS2.HWND_DESKTOP, hwnd, Message, NIL,
                                      BoxID, OS2.MB_OK + OS2.MB_ERROR);
                           OS2.WinSetFocus (OS2.HWND_DESKTOP, OS2.WinWindowFromID(hwnd, DID.UserName));
                           RETURN NIL;

                       ELSIF User.OldName[0] <> Nul THEN

                           StrToBufferA (ourlang, "Edituser.willdelete", User.OldName, Message);
                           StrToBuffer (ourlang, "Edituser.warning", stringval);
                           IF OS2.WinMessageBox (OS2.HWND_DESKTOP, hwnd, Message, stringval,
                                      BoxID, OS2.MB_OKCANCEL + OS2.MB_INFORMATION) = OS2.MBID_OK THEN
                               fhini := OpenINIForUser (User.OldName);
                               INIDeleteApp (fhini, User.OldName);
                               CloseINIFiles (fhini);
                           ELSE
                               RETURN NIL;
                           END (*IF*);

                       END (*IF*);

                       CommitChanges;
                       OS2.WinPostMsg (hwnd, OS2.WM_CLOSE, NIL, NIL);
                       RETURN NIL;

                     | DID.UserCancel:
                       User.NameChanged := FALSE;
                       OS2.WinPostMsg (hwnd, OS2.WM_CLOSE, NIL, NIL);
                       RETURN NIL;

                   ELSE
                       RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                   END (*CASE*);

          |  OS2.WM_CONTROL:

                   NotificationCode := OS2.ULONGFROMMP(mp1);
                   ButtonID := NotificationCode MOD 65536;
                   NotificationCode := NotificationCode DIV 65536;
                   ButtonValue := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, ButtonID,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;

                   IF (ButtonID = DID.UseTemplate)
                                AND (ButtonValue <> User.UseTemplate) THEN

                       EnableUseTemplateFields (hwnd, ButtonValue);
                       RETURN NIL;

                   ELSIF ButtonID = DID.UseFTP THEN
                       User.useFTP := ButtonValue;
                       EnableFields (hwnd);

                   ELSIF ButtonID = DID.UseSFTP THEN
                       User.useSFTP := ButtonValue;
                       EnableFields (hwnd);

                   ELSIF ButtonID = DID.LimitLogins THEN
                       SetLoginLimit (hwnd, ORD(ButtonValue));

                   ELSIF ButtonID = DID.limit_delete THEN
                       User.DeleteExpired := TRUE;
                       SetLoginLimit (hwnd, User.LoginLimit);

                   ELSIF ButtonID = DID.limit_deactivate THEN
                       User.DeleteExpired := FALSE;
                       SetLoginLimit (hwnd, User.LoginLimit);

                   ELSIF (ButtonID = DID.LoginLimit) AND DetectLLChanges
                                    AND (NotificationCode = OS2.EN_CHANGE) THEN

                       (* Precaution: if the field doesn't start with a digit,  *)
                       (* don't change the login limit just yet.                *)

                       OS2.WinQueryDlgItemText (hwnd, DID.LoginLimit, 8, Message);
                       IF Message[0] IN Digits THEN
                           WinQueryDlgItemCard (hwnd, DID.LoginLimit, val);
                           SetLoginLimit (hwnd, val);
                       END (*IF*);

                   ELSIF (NotificationCode = OS2.BN_CLICKED) AND ButtonValue THEN

                       PasswordWasDisabled := (User.category = NoPasswordNeeded)
                                              OR (User.category = GuestUser)
                                              OR (User.category = UserTemplate);
                       EnableIsTemplateFields (hwnd, ButtonID = DID.IsTemplate);
                       CASE ButtonID OF

                         | DID.NoPassword:
                           User.category := NoPasswordNeeded;
                           IF NOT PasswordWasDisabled THEN
                               OS2.WinQueryWindowText (passwindow, PasswordSize+1, User.Password);
                           END (*IF*);
                           OS2.WinSetWindowText (passwindow, "");
                           OS2.WinEnableWindow (passwindow, FALSE);
                           OS2.WinShowWindow (encryptwindow, FALSE);
                           OS2.WinShowWindow (allow3way, FALSE);
                           EnableFields (hwnd);
                           RETURN NIL;

                         | DID.Guest:
                           User.category := GuestUser;
                           IF NOT PasswordWasDisabled THEN
                               OS2.WinQueryWindowText (passwindow, PasswordSize+1, User.Password);
                           END (*IF*);
                           OS2.WinSetWindowText (passwindow, "@");
                           OS2.WinEnableWindow (passwindow, FALSE);
                           OS2.WinShowWindow (encryptwindow, FALSE);
                           OS2.WinShowWindow (allow3way, FALSE);
                           EnableFields (hwnd);
                           RETURN NIL;

                         | DID.User:
                           User.category := NormalUser;
                           IF PasswordWasDisabled THEN
                               OS2.WinEnableWindow (passwindow, TRUE);
                               OS2.WinShowWindow (encryptwindow, TRUE);
                               OS2.WinSetWindowText (passwindow, User.Password);
                           END (*IF*);
                           OS2.WinShowWindow (allow3way, TRUE);
                           EnableFields (hwnd);
                           RETURN NIL;

                         | DID.Manager:
                           User.category := Manager;
                           IF PasswordWasDisabled THEN
                               OS2.WinEnableWindow (passwindow, TRUE);
                               OS2.WinShowWindow (encryptwindow, TRUE);
                               OS2.WinSetWindowText (passwindow, User.Password);
                           END (*IF*);
                           OS2.WinShowWindow (allow3way, TRUE);
                           EnableFields (hwnd);
                           RETURN NIL;

                         | DID.IsTemplate:
                           User.category := UserTemplate;
                           IF NOT PasswordWasDisabled THEN
                               OS2.WinQueryWindowText (passwindow, PasswordSize+1, User.Password);
                           END (*IF*);
                           OS2.WinSetWindowText (passwindow, "");
                           OS2.WinEnableWindow (passwindow, FALSE);
                           OS2.WinShowWindow (encryptwindow, FALSE);
                           OS2.WinShowWindow (allow3way, FALSE);
                           EnableFields (hwnd);
                           RETURN NIL;

                       ELSE
                           EnableFields (hwnd);
                           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                       END (*CASE*);

                   ELSIF (ButtonID = DID.TemplateName)
                               AND (NotificationCode = OS2.EN_CHANGE) THEN

                       (* The template name has changed. *)

                       OS2.WinQueryDlgItemText (hwnd, DID.TemplateName, NameLength, User.TemplateName);
                       Tree.DeleteTree (User.TemplateTreeRoot);
                       IF User.TemplateName[0] <> Nul THEN
                           fhini := OpenINIForUser (User.TemplateName);
                           User.TemplateTreeRoot := Tree.LoadPermissions (fhini, User.TemplateName);
                           CloseINIFiles (fhini);
                       END (*IF*);
                       Tree.InitialDisplay (OS2.WinWindowFromID(hwnd, DID.DirSummary),
                                            User.TemplateTreeRoot);
                       RETURN NIL;

                   ELSIF (ButtonID = DID.DirSummary) AND NOT User.UseTemplate
                               AND (NotificationCode = OS2.LN_ENTER) THEN
                       (* Treat this one as if the "edit directory" button had been clicked. *)
                       OS2.WinSendMsg (hwnd, OS2.WM_COMMAND,
                             OS2.MPFROMSHORT(DID.EditDirButton), NIL);
                       RETURN NIL;

                   END (*IF*);
                   RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);

           |  OS2.WM_CLOSE:
                   RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);

        ELSE    (* default *)
           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
        END (*CASE*);

    END DialogueProc;

(************************************************************************)

PROCEDURE Edit (owner: OS2.HWND;  VAR (*INOUT*) name: ARRAY OF CHAR;
                            lang: LangHandle;  clone: BOOLEAN): BOOLEAN;

    (* Edit the properties of the alias called "name".  *)
    (* Returns TRUE iff the name has changed.           *)

    VAR hwnd: OS2.HWND;

    BEGIN
        OS2.DosError (OS2.FERR_DISABLEHARDERR);
        DetectLLChanges := FALSE;
        IF clone THEN
            User.OldName := "";
        ELSE
            Strings.Assign (name, User.OldName);
        END (*IF*);
        Strings.Assign (name, User.name);
        hwnd := OS2.WinLoadDlg(OS2.HWND_DESKTOP, owner,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.EditUserDialogue,                (* dialogue ID *)
                       NIL);               (* creation parameters *)
        pagehandle := hwnd;
        SetLanguage (hwnd, lang);
        CommonSettings.SetInitialWindowPosition (hwnd, "EditUser");

        LoadDialogueData (hwnd, name, lang, clone);
        EnableFields (hwnd);
        OS2.WinSetFocus (OS2.HWND_DESKTOP, OS2.WinWindowFromID(hwnd, DID.UserName));

        OS2.WinProcessDlg(hwnd);

        Strings.Assign (User.name, name);
        CommonSettings.StoreWindowPosition (hwnd, "EditUser", TRUE);
        OS2.WinDestroyWindow (hwnd);
        Tree.DeleteTree (User.UserTreeRoot);
        Tree.DeleteTree (User.TemplateTreeRoot);
        RETURN User.NameChanged;

    END Edit;

(************************************************************************)

END EditUser.

