(**************************************************************************)
(*                                                                        *)
(*  Setup for SFTPServer and FtpServer                                    *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FSUINI;

        (************************************************************)
        (*                                                          *)
        (*            PM Setup for SFtpServer/FtpServer             *)
        (*   This module opens and closes the server INI files      *)
        (*   The actual INI file operations are done by RINIData    *)
        (*                                                          *)
        (* The special feature of this version is that it can work  *)
        (* on two INI files in parallel.  In the case where both    *)
        (* files are active, we read data from the first one, but   *)
        (* write to both.                                           *)
        (*                                                          *)
        (*      Started:        27 April 2022                       *)
        (*      Last edited:    9 September 2022                    *)
        (*      Status:         OK                                  *)
        (*                                                          *)
        (************************************************************)


FROM SYSTEM IMPORT CAST, CARD8, LOC;

IMPORT RINIData, Strings, OS2;

FROM FileOps IMPORT
    (* type *)  FilenameString;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL, StringMatch;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);
    NILrhini = CAST (RINIData.RHINI, NIL);

TYPE
    FHINI = POINTER TO
                RECORD
                    rhini0, rhini1: RINIData.RHINI;
                END (*RECORD*);

VAR
    OriginalRelINIname, RelINIname, Basename, SearchMask:
                                          ARRAY [0..1] OF FilenameString;
    HashMax: CARDINAL;

    (* conflict is TRUE in the special case where two (or more) remote  *)
    (* INI/TNI files can be open at the same time.                      *)

    conflict: BOOLEAN;

    (* The directories and names of the two INI files.  *)

    Directory: ARRAY [0..1] OF FilenameString;

    INIFile: ARRAY [0..1] OF
                RECORD
                    enabled, remote: BOOLEAN;
                    relname: FilenameString;
                    fullname: FilenameString;
                END (*RECORD*);

    (* Question: do we need to store the relname field? *)

(************************************************************************)

(*
PROCEDURE DumpFileInfo;

    (* For debugging: write out the INIFile array. *)

    VAR j: [0..1];

    BEGIN
        FOR j := 0 TO 1 DO
            WITH INIFile[j] DO
                IF enabled THEN Debug.WriteString("E,") END (*IF*);
                IF remote THEN Debug.WriteString("R,") END (*IF*);
                Debug.WriteString(name);
                Debug.WriteLn;
            END (*WITH*);
        END (*FOR*);
    END DumpFileInfo;
*)

(************************************************************************)

PROCEDURE SetTNIMode (j: CARDINAL;  TNImode: BOOLEAN);

    (* Specifies whether the INI file name is FTPD.INI or FTPD.TNI.  *)

    BEGIN
        Strings.Assign ("FTPD", RelINIname[j]);
        Strings.Assign ("FTPD", Basename[j]);
        Strings.Assign ("FTPD*", SearchMask[j]);
        IF TNImode THEN
            Strings.Append (".TNI", RelINIname[j]);
            Strings.Append (".TNI", SearchMask[j]);
            Strings.Append ("    .TNI", Basename[j]);
        ELSE
            Strings.Append (".INI", RelINIname[j]);
            Strings.Append (".INI", SearchMask[j]);
            Strings.Append ("    .INI", Basename[j]);
        END (*IF*);
        Strings.Assign (RelINIname[j], OriginalRelINIname[j]);
    END SetTNIMode;

(************************************************************************)

PROCEDURE GetFullINIname (j: CARDINAL;  VAR (*OUT*) fullname: ARRAY OF CHAR;
                                    VAR (*OUT*) remote: BOOLEAN);

    (* Returns the full name of INI file j, where j is 0 or 1. If that  *)
    (* file is disabled, we return an empty string as the fullname result.*)

    BEGIN
        IF INIFile[j].enabled THEN
            Strings.Assign (INIFile[j].fullname, fullname);
            remote := INIFile[j].remote;
        ELSE
            fullname[0] := Nul;
            remote := FALSE;
        END (*IF*);
    END GetFullINIname;

(************************************************************************)

PROCEDURE SetName (dir, relname: ARRAY OF CHAR;
                                    VAR (*OUT*) INIname: ARRAY OF CHAR);

    (* Produces a name that is effectively dir\relname.  *)

    BEGIN
        Strings.Assign (dir, INIname);
        IF (dir[0] <> CHR(0)) AND (dir[LENGTH(dir)-1] <> '/')
                                AND (dir[LENGTH(dir)-1] <> '\') THEN
            Strings.Append ('\', INIname);
        END (*IF*);
        Strings.Append (relname, INIname);
    END SetName;

(************************************************************************)

PROCEDURE INInameMask (j: CARDINAL;  VAR (*OUT*) mask: ARRAY OF CHAR);

    (* Returns a search mask for finding all possible names of  *)
    (* INI file j, where j is 0 or 1.                           *)

    BEGIN
        IF INIFile[j].enabled THEN
            SetName (Directory[j], SearchMask[j], mask);
        ELSE
            mask[0] := Nul;
        END (*IF*);
    END INInameMask;

(************************************************************************)

PROCEDURE SetDirs (server: CARD8;  dir1, dir2: ARRAY OF CHAR;
                            useTNI, TNIexplicit: BOOLEAN);

    (* This must be called before any call to OpenINIFiles.         *)
    (* server is a 4-bit code with the meaning:                     *)
    (*              bit 0       enable dir1                         *)
    (*              bit 1       dir1 is remote                      *)
    (*              bit 2       enable dir2                         *)
    (*              bit 3       dir2 is remote                      *)
    (* The two BOOLEAN parameters are used in deciding whether the  *)
    (* two INI files (in directories dir1 and dir2) are called      *)
    (* FTPD.INI or FTPD.TNI                                         *)

    VAR xfer :  RECORD
                    CASE :BOOLEAN OF
                        FALSE:  cval: CARD8;
                      |
                        TRUE:   bval: PACKEDSET OF [0..3];
                    END (*CASE*);
                END (*RECORD*);
        j: CARDINAL;
        useTNI0: BOOLEAN;

    BEGIN
        conflict := server = 0FH;
        xfer.cval := server;
        INIFile[0].enabled := 0 IN xfer.bval;
        INIFile[0].remote  := 1 IN xfer.bval;
        INIFile[1].enabled := 2 IN xfer.bval;
        INIFile[1].remote  := 3 IN xfer.bval;
        Strings.Assign (dir1, Directory[0]);
        Strings.Assign (dir2, Directory[1]);

        (* If only one of the two is enabled, move that to position 0.  *)
        (* Note that it doesn't matter if we corrupt the details of     *)
        (* the non-enabled one.                                         *)

        IF NOT INIFile[0].enabled THEN
            INIFile[0] := INIFile[1];
            Directory[0] := Directory[1];
            INIFile[1].enabled := FALSE;
        END (*IF*);

        (* Now that we've settled the local/remote question, we can *)
        (* finally make decisions about INI vs TNI.                 *)

        useTNI0 := useTNI;
        FOR j := 0 TO 1 DO
            IF INIFile[j].enabled THEN
                useTNI := useTNI0;
                RINIData.DecideTNImode (Directory[j], "FTPD",
                                TNIexplicit, INIFile[j].remote, useTNI);
                SetTNIMode (j, useTNI);
                INIFile[j].relname := RelINIname[j];
                SetName (Directory[j], RelINIname[j], INIFile[j].fullname);
            END (*IF*);
        END (*FOR*);

        IF INIFile[0].enabled AND INIFile[1].enabled THEN

            (* Both are enabled, but cut this back to one if the two    *)
            (* file names are the same.                                 *)

            IF StringMatch (INIFile[0].fullname, INIFile[1].fullname) THEN
                INIFile[1].enabled := FALSE;
            END (*IF*);
        END (*IF*);

    END SetDirs;

(************************************************************************)

PROCEDURE RemoteConflict (): BOOLEAN;

    (* Returns TRUE iff both INI files are remote. *)

    BEGIN
        RETURN conflict;
    END RemoteConflict;

(************************************************************************)

PROCEDURE ChangeINIFilename (name: ARRAY OF CHAR);

    (* Temporarily changes the specification of the INI file name. *)

    BEGIN
        Strings.Assign (name, RelINIname[0]);
    END ChangeINIFilename;

(************************************************************************)

PROCEDURE RestoreINIFilename;

    (* Sets the INI file name back to its original value. *)

    BEGIN
        Strings.Assign (OriginalRelINIname[0], RelINIname[0]);
    END RestoreINIFilename;

(************************************************************************)

PROCEDURE OpenINIFiles (): FHINI;

    (* Opens one or both main INI files. (If the users are spread out   *)
    (* over multiple INI files, that is handled by a different proc.)   *)

    VAR result: FHINI;

    BEGIN
        NEW (result);
        IF INIFile[0].enabled THEN
            result^.rhini0 := RINIData.OpenINIFile (INIFile[0].fullname,
                                            INIFile[0].remote, conflict);
        ELSE
            result^.rhini0 := NIL;
        END (*IF*);
        IF INIFile[1].enabled THEN
            result^.rhini1 := RINIData.OpenINIFile (INIFile[1].fullname,
                                            INIFile[1].remote, conflict);
        ELSE
            result^.rhini1 := NIL;
        END (*IF*);
        RETURN result;
    END OpenINIFiles;

(************************************************************************)

PROCEDURE CloseINIFiles (fhini: FHINI);

    BEGIN
        RINIData.CloseINIFile (fhini^.rhini0);
        RINIData.CloseINIFile (fhini^.rhini1);
        DISPOSE (fhini);
    END CloseINIFiles;

(************************************************************************)

PROCEDURE FirstIsRemote (fhini: FHINI): BOOLEAN;

    (* Returns TRUE iff we are going to read from a remote file. *)

    BEGIN
        RETURN INIFile[0].enabled AND INIFile[0].remote;
    END FirstIsRemote;

(************************************************************************)
(*          PROVISION FOR SPREADING USERS AMONG MULTIPLE INI FILES      *)
(************************************************************************)

PROCEDURE SetHashMax (value: CARDINAL);

    (* Sets the number of extra INI files to be used for the "large     *)
    (* number of users" case.  If value=0 then we revert to the default *)
    (* of using a single INI file.  Otherwise, it is recommended that   *)
    (* value be a prime number.                                         *)

    BEGIN
        HashMax := value;
    END SetHashMax;

(************************************************************************)

PROCEDURE GetHashMax (j: CARDINAL): CARDINAL;

    (* Returns the HashMax value stored in INI file j.  *)

    VAR rhini: RINIData.RHINI;
        result: CARDINAL;

    BEGIN
        IF NOT INIFile[0].enabled THEN RETURN 0 END (*IF*);
        rhini := RINIData.OpenINIFile (INIFile[j].fullname,
                                            INIFile[j].remote, conflict);
        IF NOT RINIData.INIFetch (rhini, "$SYS", "HashMax", result) THEN
            HashMax := 0;
        END (*IF*);
        RINIData.CloseINIFile (rhini);
        RETURN result;
    END GetHashMax;

(************************************************************************)

PROCEDURE OpenAlternative (j: CARDINAL;  name: ARRAY OF CHAR): RINIData.RHINI;

    (* Opens an INI file 'name' that is in the same directory as INI file j. *)

    VAR filename: FilenameString;
        result: RINIData.RHINI;

    BEGIN
        SetName (Directory[j], name, filename);
        IF INIFile[j].enabled THEN
            result := RINIData.OpenINIFile (filename,
                                            INIFile[j].remote, conflict);
        ELSE
            result := NIL;
        END (*IF*);
        RETURN result;
    END OpenAlternative;

(************************************************************************)

PROCEDURE DeleteAlternative (j: CARDINAL;  name: ARRAY OF CHAR;  status: OS2.HWND);

    (* Deletes the file that was opened by OpenAlternative.  *)

    VAR filename: FilenameString;

    BEGIN
        SetName (Directory[j], name, filename);
        RINIData.DeleteFile (filename, INIFile[j].remote, status);
    END DeleteAlternative;

(************************************************************************)

PROCEDURE OpenRINIForUser (j: CARDINAL;  name: ARRAY OF CHAR): RINIData.RHINI;

    (* Opens the INI file that contains the data for the named user. The*)
    (* directory is taken from the global variable Directory, and the   *)
    (* enabled and remote flags are copied from global variable INIFile.*)

    VAR relname: FilenameString;
        INIname: FilenameString;
        code, k, length: CARDINAL;
        result: RINIData.RHINI;

    BEGIN
        IF HashMax = 0 THEN
            INIname := INIFile[j].fullname;
        ELSE
            Strings.Capitalize (name);

            (* Compute a hash code from the name. *)

            code := 0;
            length := Strings.Length(name);
            IF length > 0 THEN
                FOR k := 0 TO length-1 DO
                    code := (16*code + ORD(name[k])) MOD HashMax;
                END (*FOR*);
            END (*IF*);

            (* Turn the code into a file name. *)

            IF INIFile[j].enabled THEN
                relname := Basename[j];
                FOR k := 7 TO 4 BY -1 DO
                    relname[k] := CHR(code MOD 10 + ORD('0'));
                    code := code DIV 10;
                END (*FOR*);

                SetName (Directory[j], relname, INIname);
            END (*IF*);

        END (*IF*);

        IF INIFile[j].enabled THEN
            result := RINIData.OpenINIFile (INIname,
                                            INIFile[j].remote, conflict);
        ELSE
            result := NIL;
        END (*IF*);
        RETURN result;

    END OpenRINIForUser;

(************************************************************************)

PROCEDURE OpenINIForUser (name: ARRAY OF CHAR): FHINI;

    (* Opens the INI file that contains the data for the named user.    *)
    (* Or possibly two INI files.  The directories are taken from the   *)
    (* global variable Directory, and the enabled and remote flags      *)
    (* are copied from global variable INIFile.                         *)

    VAR relname: FilenameString;
        INIname: ARRAY [0..1] OF FilenameString;
        code, j, k, length: CARDINAL;
        result: FHINI;

    BEGIN
        NEW (result);
        IF HashMax = 0 THEN
            INIname[0] := INIFile[0].fullname;
            INIname[1] := INIFile[1].fullname;
        ELSE
            Strings.Capitalize (name);

            (* Compute a hash code from the name. *)

            code := 0;
            length := Strings.Length(name);
            IF length > 0 THEN
                FOR k := 0 TO length-1 DO
                    code := (16*code + ORD(name[k])) MOD HashMax;
                END (*FOR*);
            END (*IF*);

            (* Turn the code into a file name. *)

            FOR j := 0 TO 1 DO
                IF INIFile[j].enabled THEN
                    relname := Basename[j];
                    FOR k := 7 TO 4 BY -1 DO
                        relname[k] := CHR(code MOD 10 + ORD('0'));
                        code := code DIV 10;
                    END (*FOR*);

                    SetName (Directory[j], relname, INIname[j]);
                END (*IF*);
            END (*FOR*);

        END (*IF*);

        IF INIFile[0].enabled THEN
            result^.rhini0 := RINIData.OpenINIFile (INIname[0],
                                            INIFile[0].remote, conflict);
        ELSE
            result^.rhini0 := NIL;
        END (*IF*);
        IF INIFile[1].enabled THEN
            result^.rhini1 := RINIData.OpenINIFile (INIname[1],
                                            INIFile[1].remote, conflict);
        ELSE
            result^.rhini1 := NIL;
        END (*IF*);
        RETURN result;

    END OpenINIForUser;

(************************************************************************)
(*                           LOOKUP FUNCTIONS                           *)
(************************************************************************)

PROCEDURE ItemSize (fhini: FHINI;  app, key: ARRAY OF CHAR;
                                        VAR (*OUT*) size: CARDINAL): BOOLEAN;

    (* Returns the number of bytes in the specified INI item.  If the item does *)
    (* not exist, returns FALSE with size set to zero.                          *)

    BEGIN
        RETURN RINIData.ItemSize (fhini^.rhini0, app, key, size);
    END ItemSize;

(************************************************************************)

PROCEDURE INIFetch (fhini: FHINI;  app, key: ARRAY OF CHAR;
                    VAR (*OUT*) variable: ARRAY OF LOC): BOOLEAN;

    (* Reads data from the INI file. *)

    BEGIN
        RETURN RINIData.INIFetch (fhini^.rhini0, app, key, variable);
    END INIFetch;

(************************************************************************)

PROCEDURE INIFetchBinary (fhini: FHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) result: ARRAY OF LOC;
                                            size: CARDINAL): BOOLEAN;

    (* Retrieves the value of a variable from the INI file.  Returns FALSE if   *)
    (* the variable was not found.  This the version in which we trust the      *)
    (* caller to have ensured that the size is correct.                         *)

    BEGIN
        RETURN RINIData.INIFetchBinary (fhini^.rhini0, app, key, result, size);
    END INIFetchBinary;

(************************************************************************)

PROCEDURE INIGetCard (fhini: FHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) val: CARDINAL): BOOLEAN;

    (* Gets a single cardinal value from the INI file, returns true if OK. *)

    BEGIN
        RETURN RINIData.INIGetCard (fhini^.rhini0, app, key, val);
    END INIGetCard;

(************************************************************************)

PROCEDURE INIGetString (fhini: FHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) val: ARRAY OF CHAR): BOOLEAN;

    (* Gets a string value from the INI file, returns true if OK. *)

    BEGIN
        RETURN RINIData.INIGetString (fhini^.rhini0, app, key, val);
    END INIGetString;

(************************************************************************)

PROCEDURE GetStringList (fhini: FHINI;  app, key: ARRAY OF CHAR;
                                   VAR (*OUT*) state: StringReadState);

    (* Initialisation in preparation for a "NextString" operation. *)

    BEGIN
        RINIData.GetStringList (fhini^.rhini0, app, key, state);
    END GetStringList;

(************************************************************************)

PROCEDURE NextString (state: StringReadState;  VAR (*OUT*) result: ARRAY OF CHAR);

    (* Reads the next character string from a string-of-strings field.  *)
    (* An empty string is returned when we have run out of strings.     *)

    BEGIN
        RINIData.NextString (state, result);
    END NextString;

(************************************************************************)

PROCEDURE CloseStringList (VAR (*INOUT*) state: StringReadState);

    (* Must be called to release the memory used in fetching a  *)
    (* string of strings.                                       *)

    BEGIN
        RINIData.CloseStringList (state);
    END CloseStringList;

(************************************************************************)
(*                             WRITING DATA                             *)
(************************************************************************)

PROCEDURE INIPut (fhini: FHINI;  app, key: ARRAY OF CHAR;  VAR (*IN*) value: ARRAY OF LOC);

    (* Writes data to the INI file. *)

    VAR hini: RINIData.RHINI;

    BEGIN
        hini := fhini^.rhini0;
        IF hini <> NILrhini THEN
            RINIData.INIPut (hini, app, key, value);
        END (*IF*);
        hini := fhini^.rhini1;
        IF hini <> NILrhini THEN
            RINIData.INIPut (hini, app, key, value);
        END (*IF*);
    END INIPut;

(************************************************************************)

PROCEDURE INIPutBinary (fhini: FHINI;  app, key: ARRAY OF CHAR;
                    VAR (*IN*) variable: ARRAY OF LOC;  NumberOfBytes: CARDINAL);

    (* Writes a binary string of the specified length to the INI file. *)

    VAR hini: RINIData.RHINI;

    BEGIN
        hini := fhini^.rhini0;
        IF hini <> NILrhini THEN
            RINIData.INIPutBinary (hini, app, key, variable, NumberOfBytes);
        END (*IF*);
        hini := fhini^.rhini1;
        IF hini <> NILrhini THEN
            RINIData.INIPutBinary (hini, app, key, variable, NumberOfBytes);
        END (*IF*);
    END INIPutBinary;

(************************************************************************)

PROCEDURE INIPutString (fhini: FHINI;  app, key: ARRAY OF CHAR;
                           VAR (*IN*) variable: ARRAY OF CHAR);

    (* Writes a character string to the INI file. *)

    VAR hini: RINIData.RHINI;

    BEGIN
        hini := fhini^.rhini0;
        IF hini <> NILrhini THEN
            RINIData.INIPutString (hini, app, key, variable);
        END (*IF*);
        hini := fhini^.rhini1;
        IF hini <> NILrhini THEN
            RINIData.INIPutString (hini, app, key, variable);
        END (*IF*);
    END INIPutString;

(************************************************************************)
(*                             MISCELLANEOUS                            *)
(************************************************************************)

PROCEDURE INIDeleteApp (fhini: FHINI;  app: ARRAY OF CHAR);

    (* Deletes an application from both INI files. *)

    BEGIN
        IF fhini^.rhini0 <> NILrhini THEN
            RINIData.INIDeleteApp (fhini^.rhini0, app);
        END (*IF*);
        IF fhini^.rhini1 <> NILrhini THEN
            RINIData.INIDeleteApp (fhini^.rhini1, app);
        END (*IF*);
    END INIDeleteApp;

(************************************************************************)

PROCEDURE INIDeleteKey (fhini: FHINI;  app, key: ARRAY OF CHAR);

    (* Deletes a key from both INI files. *)

    VAR hini: RINIData.RHINI;

    BEGIN
        hini := fhini^.rhini0;
        IF hini <> NILrhini THEN
            RINIData.INIDeleteKey (hini, app, key);
        END (*IF*);
        hini := fhini^.rhini1;
        IF hini <> NILrhini THEN
            RINIData.INIDeleteKey (hini, app, key);
        END (*IF*);
    END INIDeleteKey;

(************************************************************************)
(*                            INITIALISATION                            *)
(************************************************************************)

BEGIN
    SetTNIMode (0, FALSE);
    SetTNIMode (1, FALSE);
    HashMax := 0;
END FSUINI.

